class Camato < Gtk::Window
  attr_reader :map, :n_rb, :n_mono, :n_plenty, :n_chapel, :n_univ,
  :n_governor, :n_library, :n_market, :n_soldier

  def initialize(n_row, n_col)
    super Gtk::Window::Type::TOPLEVEL
    @buttons = Hash.new
    set_title(_("Untitled - Camato"))

    @tt = Gtk::Tooltips.new
    @tt.enable

    vbox = Gtk::VBox.new(false, 0)
    hbox = Gtk::HBox.new(false, 0)

    # Build the menu bar
    ag = Gtk::AccelGroup.new
    add_accel_group(ag)

    mb = Gtk::MenuBar.new
    # Menu -> File
    mi = Gtk::MenuItem.new(_('_File'))
    mb.append(mi)
    menu = Gtk::Menu.new
    mi.set_submenu(menu)
    mi = Gtk::TearoffMenuItem.new
    menu.append(mi)
    # Menu -> File -> New
    mi = Gtk::ImageMenuItem.new(Gtk::Stock::NEW, ag)
    mi.signal_connect('activate') {
      d = Gtk::Dialog.new
      d.set_title(_('New map'))

      table = Gtk::Table.new(2, 2, false)
      table.set_row_spacings(5)
      table.set_column_spacings(4)
      table.attach(Gtk::Label.new(_('Width')), 0,1, 0,1)
      table.attach(Gtk::Label.new(_('Height')), 0,1, 1,2)
      table.attach(Gtk::Label.new(_('Title')), 0,1, 2,3)
      new_width = Gtk::SpinButton.new(4,100,1)
      new_width.set_value(8)
      new_height = Gtk::SpinButton.new(4,100,1)
      new_height.set_value(7)
      new_title = Gtk::Entry.new
      table.attach(new_width, 1,2, 0,1)
      table.attach(new_height, 1,2, 1,2)
      table.attach(new_title, 1,2, 2,3)
      d.vbox.pack_start(table, false, true, 0)

      bb = Gtk::HButtonBox.new
      ok = Gtk::Button.new(Gtk::Stock::OK)
      ok.signal_connect('clicked') {
        new_map(new_height.text.to_i,
                new_width.text.to_i)
        @title.set_text(new_title.text)
        d.hide
      }
      cancel = Gtk::Button.new(Gtk::Stock::CANCEL)
      cancel.signal_connect('clicked') {
        d.hide
      }
      bb.pack_start(cancel)
      bb.pack_start(ok)

      d.action_area.pack_start(bb, false, true, 0)
      d.show_all
    }
    menu.append(mi)

    # Menu -> File -> Open
    mi = Gtk::ImageMenuItem.new(Gtk::Stock::OPEN, ag)
    mi.signal_connect('activate') {
      fs = Gtk::FileSelection.new(_('Open'))
      fs.set_filename(@wd) unless @wd == nil
      fs.ok_button.signal_connect('clicked') {
	fs.hide
	begin
	  open fs.filename
	rescue
	  m = message_box($!.to_s)
	end
      }
      fs.cancel_button.signal_connect('clicked') {
	fs.hide
      }
      fs.show
    }
    menu.append(mi)

    # Menu -> File -> Save
    mi = Gtk::ImageMenuItem.new(Gtk::Stock::SAVE, ag)
    mi.signal_connect('activate') {
      begin
        if @file == nil
          save_as
        else
          save @file
        end
      rescue
        m = message_box($!.to_s)
      end
    }
    menu.append(mi)
    mi = Gtk::ImageMenuItem.new(Gtk::Stock::SAVE_AS, ag)
    mi.signal_connect('activate') {
      save_as
    }
    menu.add mi
    mi = Gtk::MenuItem.new
    menu.add mi
    mi = Gtk::ImageMenuItem.new(Gtk::Stock::QUIT, ag)
    mi.signal_connect('activate') { exit }
    menu.add mi

    # Menu -> Tools
    menu = Gtk::MenuItem.new _('_Tools')
    tools = Gtk::Menu.new
    mi = Gtk::TearoffMenuItem.new
    tools.add mi
    mi = Gtk::ImageMenuItem.new _('_Add row or column')
    image = Gtk::Image.new
    image.set(Gtk::Stock::ADD, Gtk::IconSize::MENU)
    mi.set_image(image)
    mi.signal_connect('activate') {
      select_row_col(_('Add row or column')) { |i|
	case i
	when 0
	  a = Array.new
	  0.upto(@map.width - 1) { a.push Hex.new('sea') }
	  @map.hex.unshift a
	  @map.height = @map.height + 1
	  # Correct the shift of even rows	  
	  0.upto(@map.height - 1) { |i|
	    if (i.modulo(2) == 1)
	      @map.hex[i].shift
	      @map.hex[i].push Hex.new('sea')
	    end
	  }
	when 1
	  @map.hex.each { |a| a.push Hex.new('sea') }
	  @map.width = @map.width + 1
	when 2
	  a = Array.new
	  0.upto(@map.width - 1) { a.push Hex.new('sea') }
	  @map.hex.push a
	  @map.height = @map.height + 1
	when 3
	  @map.hex.each { |a| a.unshift Hex.new('sea') }
	  @map.width = @map.width + 1
	end
      
	@map.repaint
	@map.redisplay
      }
    }
    tools.add mi
    mi = Gtk::ImageMenuItem.new _('_Remove row or column')
    image = Gtk::Image.new
    image.set(Gtk::Stock::REMOVE, Gtk::IconSize::MENU)
    mi.set_image(image)
    mi.signal_connect('activate') {
      select_row_col(_('Remove row or column')) { |i|
	case i
	when 0
	  @map.hex.shift
	  @map.height = @map.height - 1 unless @map.height == 0
	  # Correct the shift of even rows	  
	  0.upto(@map.height - 1) { |i|
	    if (i.modulo(2) == 1)
	      @map.hex[i].shift
	      @map.hex[i].push Hex.new('sea')
	    end
	  }
	when 1
	  @map.hex.each { |array| array.pop }
	  @map.width = @map.width - 1 unless @map.width == 0
	when 2
	  @map.hex.pop
	  @map.height = @map.height - 1 unless @map.height == 0
	when 3
	  @map.hex.each { |array| array.shift }
	  @map.width = @map.width - 1 unless @map.width == 0
	end
      
	@map.repaint
	@map.redisplay
      }
    }
    tools.add mi
    mi = Gtk::MenuItem.new _('Remove all _no-setup nodes')
    mi.signal_connect('activate') {
      @map.ns = NoSetupNodes.new
      @map.repaint
      @map.redisplay
    }
    tools.add mi
    mi = Gtk::MenuItem.new
    tools.add mi
    mi = Gtk::MenuItem.new _('_Statistics')
    mi.signal_connect('activate') {
      a = Stats.new(self)
      a.show_all
    }
    tools.add mi
    menu.set_submenu tools
    mb.append menu

    # Menu -> Help
    menu = Gtk::MenuItem.new _('_Help')
    help = Gtk::Menu.new
    mi = Gtk::TearoffMenuItem.new
    help.add mi
    mi = Gtk::MenuItem.new _('_Keyboard shortcuts')
    mi.signal_connect('activate') {
      d = Gtk::Dialog.new
      d.set_title(_('Keyboard shortcuts'))
      l = Gtk::ListStore.new(String, String)
      i = l.append
      i[0] = 'A'
      i[1] = _('Gold (Any)')
      i = l.append
      i[0] = 'B'
      i[1] = _('Brick')
      i = l.append
      i[0] = 'D'
      i[1] = _('Desert')
      i = l.append
      i[0] = 'G'
      i[1] = _('Grain')
      i = l.append
      i[0] = 'L'
      i[1] = _('Lumber')
      i = l.append
      i[0] = 'O'
      i[1] = _('Ore')
      i = l.append
      i[0] = 'P'
      i[1] = _('Port (3:1)')
      i = l.append
      i[0] = 'S'
      i[1] = _('Sea')
      i = l.append
      i[0] = 'V'
      i[1] = _('Void')
      i = l.append
      i[0] = 'W'
      i[1] = _('Wool')
      i = l.append
      i[0] = 'N'
      i[1] = _('No-setup node')
      i = l.append
      i[0] = 'H'
      i[1] = _('No-setup hexagon')
      if $use_cvs
        i = l.append
        i[0] = 'R'
        i[1] = _('Don\'t randomize hexagon')
      end
      i = l.append
      i[0] = _('<space>')
      i[1] = _('Rotate port')
      view = Gtk::TreeView.new(l)
      renderer = Gtk::CellRendererText.new()
      col = Gtk::TreeViewColumn.new(_('Key'), renderer, :text => 0)
      view.append_column(col)
      col = Gtk::TreeViewColumn.new(_('Action'), renderer, :text => 1)
      view.append_column(col)

      s = Gtk::ScrolledWindow.new
      s.add(view)
      s.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
      s.set_size_request(250, 200)
      d.vbox.pack_start(s)
      close = Gtk::Button.new(Gtk::Stock::CLOSE)
      close.signal_connect('clicked') { d.hide }
      d.action_area.pack_start close
      d.show_all
    }
    help.add mi
    mi = Gtk::MenuItem.new _('_About Camato')
    mi.signal_connect('activate') {
      str = "Camato #{MY_VERSION}, (c) Yusei, 2003-2005\n<yusei@ragondux.com>\n\n"
      str << _('Camato is free software, released under the GPL.')
      message_box(str)
    }
    help.add mi
    menu.set_submenu help
    mb.append menu

    vbox.pack_start mb, false, true, 0

    # Toolbar
    tb = Gtk::Table.new(7, 2, true)
    tb.set_row_spacings 5
    tb.set_column_spacings 5

    icon = Gtk::Image.new(get_path("img/sheep.png"))
    @buttons['w'] = Gtk::ToggleButton.new
    @buttons['w'].add(icon)
    @tt.set_tip(@buttons['w'], _('Wool'), nil)
    tb.attach(@buttons['w'], 0,1,0,1)

    icon = Gtk::Image.new(get_path("img/bricks.png"))
    @buttons['b'] = Gtk::ToggleButton.new
    @buttons['b'].add(icon)
    @tt.set_tip(@buttons['b'], _('Brick'), nil)
    tb.attach(@buttons['b'], 0,1,1,2)

    icon = Gtk::Image.new(get_path("img/grain.png"))
    @buttons['g'] = Gtk::ToggleButton.new
    @buttons['g'].add(icon)
    @tt.set_tip(@buttons['g'], _('Grain'), nil)
    tb.attach(@buttons['g'], 0,1,2,3)
    
    icon = Gtk::Image.new(get_path("img/ore.png"))
    @buttons['o'] = Gtk::ToggleButton.new
    @buttons['o'].add(icon)
    @tt.set_tip(@buttons['o'], _('Ore'), nil)
    tb.attach(@buttons['o'], 0,1,3,4)

    icon = Gtk::Image.new(get_path("img/lumber.png"))
    @buttons['l'] = Gtk::ToggleButton.new
    @buttons['l'].add(icon)
    @tt.set_tip(@buttons['l'], _('Lumber'), nil)
    tb.attach(@buttons['l'], 0,1,4,5)

    icon = Gtk::Image.new(get_path("img/sea.png"))
    @buttons['s'] = Gtk::ToggleButton.new
    @buttons['s'].add(icon)
    @tt.set_tip(@buttons['s'], _('Sea'), nil)
    tb.attach(@buttons['s'], 0,1,6,7)

    icon = Gtk::Image.new(get_path("img/void.png"))
    @buttons['v'] = Gtk::ToggleButton.new
    @buttons['v'].add(icon)
    @tt.set_tip(@buttons['v'], _('Void'), nil)
    tb.attach(@buttons['v'], 0,1,5,6)

    icon = Gtk::Image.new(get_path("img/desert.png"))
    @buttons['d'] = Gtk::ToggleButton.new
    @tt.set_tip(@buttons['d'], _('Desert'), nil)
    @buttons['d'].add(icon)
    tb.attach(@buttons['d'], 1,2,6,7)

    @buttons['s?'] = Gtk::ToggleButton.new('3:1')
    @tt.set_tip(@buttons['s?'], _('Generic port'), nil)
    tb.attach(@buttons['s?'], 1,2,5,6)
    @buttons['sw'] = Gtk::ToggleButton.new('2:1 W')
    @tt.set_tip(@buttons['sw'], _('Wool port'), nil)
    tb.attach(@buttons['sw'], 1,2,0,1)
    @buttons['sb'] = Gtk::ToggleButton.new('2:1 B')
    @tt.set_tip(@buttons['sb'], _('Brick port'), nil)
    tb.attach(@buttons['sb'], 1,2,1,2)
    @buttons['sg'] = Gtk::ToggleButton.new('2:1 G')
    @tt.set_tip(@buttons['sg'], _('Grain port'), nil)
    tb.attach(@buttons['sg'], 1,2,2,3)
    @buttons['so'] = Gtk::ToggleButton.new('2:1 O')
    @tt.set_tip(@buttons['so'], _('Ore port'), nil)
    tb.attach(@buttons['so'], 1,2,3,4)
    @buttons['sl'] = Gtk::ToggleButton.new('2:1 L')
    @tt.set_tip(@buttons['sl'], _('Lumber port'), nil)
    tb.attach(@buttons['sl'], 1,2,4,5)

    icon = Gtk::Image.new(get_path("img/gold.png"))
    @buttons['a'] = Gtk::ToggleButton.new
    @buttons['a'].add(icon)
    @tt.set_tip(@buttons['a'], _('Gold'), nil)
    tb.attach(@buttons['a'], 0,1,7,8)

    @buttons['rotate'] = Gtk::ToggleButton.new(_('Rotate port'))
    tb.attach(@buttons['rotate'], 0,2,8,9)
    @buttons['nosetup'] = Gtk::ToggleButton.new(_('No-setup node'))
    tb.attach(@buttons['nosetup'], 0,2,9,10)
    @buttons['ns_hex'] = Gtk::ToggleButton.new(_('No-setup hex'))
    tb.attach(@buttons['ns_hex'], 0,2,10,11)
    if ($use_cvs) then
      @buttons['no_shuffle'] = Gtk::ToggleButton.new(_('Not random'))
      tb.attach(@buttons['no_shuffle'], 0,2,11,12)
    end
    hbox.pack_start(tb, false, true, 0)

    # Development cards
    dc = Gtk::VBox.new false, 0
    tab = Gtk::Table.new 9,2, false
    tab.row_spacings = 4

    tab.attach(Gtk::Label.new(_('Road Building')), 0,1, 0,1)
    @n_rb = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_rb, 1,2, 0,1)

    tab.attach(Gtk::Label.new(_('Monopoly')), 0,1, 1,2)
    @n_mono = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_mono, 1,2, 1,2)

    tab.attach(Gtk::Label.new(_('Year of Plenty')), 0,1, 2,3)
    @n_plenty = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_plenty, 1,2, 2,3)

    tab.attach(Gtk::Label.new(_('Chapel')), 0,1, 3,4)
    @n_chapel = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_chapel, 1,2, 3,4)

    tab.attach(Gtk::Label.new(_('University')), 0,1, 4,5)
    @n_univ = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_univ, 1,2, 4,5)

    tab.attach(Gtk::Label.new(_('Governor House')), 0,1, 5,6)
    @n_governor = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_governor, 1,2, 5,6)

    tab.attach(Gtk::Label.new(_('Library')), 0,1, 6,7)
    @n_library = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_library, 1,2, 6,7)

    tab.attach(Gtk::Label.new(_('Market')), 0,1, 7,8)
    @n_market = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_market, 1,2, 7,8)

    tab.attach(Gtk::Label.new(_('Soldiers')), 0,1, 8,9)
    @n_soldier = Gtk::SpinButton.new(0,100,1)
    tab.attach(@n_soldier, 1,2, 8,9)

    dc.pack_start tab, false, false, 0

    # Properties
    prop = Gtk::VBox.new false, 0
    tab = Gtk::Table.new 3,2, false
    tab.row_spacings = 4

    tab.attach(Gtk::Label.new(_('Title')), 0,1, 0,1)
    @title = Gtk::Entry.new
    tab.attach(@title, 1,2, 0,1)

    tab.attach(Gtk::Label.new(_('Number of players')), 0,1, 1,2)
    @nop = Gtk::SpinButton.new(2, 42, 1)
    tab.attach(@nop, 1,2, 1,2)

    tab.attach(Gtk::Label.new(_('Victory Point Traget')), 0,1, 2,3)
    @vp = Gtk::SpinButton.new(2, 42, 1)
    tab.attach(@vp, 1,2, 2,3)

    @trade = Gtk::CheckButton.new _('Interplayer trading allowed')
    tab.attach(@trade, 1,2, 3,4)

    @stricttrade = Gtk::CheckButton.new _('Trading allowed only before build')
    tab.attach(@stricttrade, 1,2, 4,5)

    tab.attach(Gtk::Label.new(_('Number of roads')), 0,1, 5,6)
    @n_roads = Gtk::SpinButton.new(2, 100, 1)
    tab.attach(@n_roads, 1,2, 5,6)

    tab.attach(Gtk::Label.new(_('Number of ships')), 0,1, 6,7)
    @n_ships = Gtk::SpinButton.new(0, 100, 1)
    tab.attach(@n_ships, 1,2, 6,7)

    tab.attach(Gtk::Label.new(_('Number of settlements')), 0,1, 7,8)
    @n_settl = Gtk::SpinButton.new(2, 100, 1)
    tab.attach(@n_settl, 1,2, 7,8)

    tab.attach(Gtk::Label.new(_('Number of cities')), 0,1, 8,9)
    @n_city = Gtk::SpinButton.new(0, 100, 1)
    tab.attach(@n_city, 1,2, 8,9)

    tab.attach(Gtk::Label.new(_('Number of bridges')), 0,1, 9,10)
    @n_bridge = Gtk::SpinButton.new(0, 100, 1)
    tab.attach(@n_bridge, 1,2, 9,10)

    tab.attach(Gtk::Label.new(_('Resource count')), 0,1, 10,11)
    @n_resource = Gtk::SpinButton.new(2, 100, 1)
    tab.attach(@n_resource, 1,2, 10,11)

    @use_pirate = Gtk::CheckButton.new _('Use pirate ship rule')
    tab.attach(@use_pirate, 1, 2, 11, 12)

    @rand_terrain = Gtk::CheckButton.new _('Random terrain')
    tab.attach(@rand_terrain, 1,2, 12,13)

    reset

    prop.pack_start tab, false, false, 0

    # Comments panel
    comm = Gtk::VBox.new(false, 4)
    warning = Gtk::HBox.new(false, 0)
    icon = Gtk::Image.new
    icon.set(Gtk::Stock::DIALOG_WARNING, Gtk::IconSize::DND)
    warning.pack_start(icon, true, true, 0)
    e = Gtk::Label.new
    e.set_text(_('Lines that are not understood by camato'+
                   " are considered as comments. Comments "+
                   "may include usefull information for " +
                   'gnocatan, so edit them at your own risk. Feel free to '+
                   'add comment at the end of the file.'))
    e.wrap = true
    warning.pack_start(e, true, true, 0)
    comm.pack_start(warning, false, true, 0)
    comm.pack_start(Gtk::HSeparator.new, false, false, 0)

    # 'Start' comments doesn't seem to be supported, so I don't provide a graphical way
    # to edit them, but I will still keep them, just in case. 
    # comm.pack_start(Gtk::Label.new(_('Beginning of the file:')), false, true, 0)
    @start_comment = Gtk::TextView.new
    # s = Gtk::ScrolledWindow.new
    # s.add(@start_comment)
    # s.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS)
    # comm.pack_start(s, true, true, 0)

    comm.pack_start(Gtk::Label.new(_('End of the file:')), false, true, 0)
    @end_comment = Gtk::TextView.new
    s = Gtk::ScrolledWindow.new
    s.add(@end_comment)
    s.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS)
    comm.pack_start(s, true, true, 0)
    comm.pack_start(Gtk::Label.new(_('Between headers and map (for unsupported commands):')),
                    false, true, 0)
    @middle_comment = Gtk::TextView.new
    s = Gtk::ScrolledWindow.new
    s.add(@middle_comment)
    s.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS)
    comm.pack_start(s, true, true, 0)

    # Tabbed Window
    nb = Gtk::Notebook.new
    mapvb = Gtk::VBox.new(false, 0)
    @map = Map.new(n_row, n_col)
    # @map = Gtk::Button.new 'test'
    @map.signal_connect('button_press_event') {
      @map.click(@tool)
    }

    mapvb.pack_start(@map, true, true, 0)

    # Number chits panel
    @chits_panel = Gtk::HBox.new(true, 0)
    [2, 3, 4, 5, 6, 8, 9, 10, 11, 12].each { |i|
      @buttons[i.to_s] = Gtk::ToggleButton.new(i.to_s)
      @chits_panel.pack_start(@buttons[i.to_s], false, true, 0)
    }
    # Attach signals to the toolbar buttons
    @buttons.each { |key, b|
      b.set_relief(Gtk::ReliefStyle::NONE)
      b.signal_connect('clicked') { |widget|
	  select_tool(key) if b.active?
      }
    }
    mapvb.pack_start(@chits_panel, false, true, 0)

    nb.append_page mapvb, Gtk::Label.new(_('Map'))
    nb.append_page prop, Gtk::Label.new(_('Properties'))
    nb.append_page dc, Gtk::Label.new(_('Development cards'))
    nb.append_page comm, Gtk::Label.new(_('Comments'))

    hbox.pack_start(nb)
    vbox.pack_start(hbox)
    add(vbox)
    show_all

#    @chits_panel.show_all

    signal_connect('destroy') { Gtk.main_quit }
  end

  def select_tool(name)
    @buttons.each { |key, val|
	val.set_active(false) unless (key === name)
    }
    @tool = name
  end

  def message_box(message)
    d = Gtk::MessageDialog.new(self,
                               Gtk::Dialog::DESTROY_WITH_PARENT,
                               Gtk::MessageDialog::INFO,
                               Gtk::MessageDialog::BUTTONS_CLOSE,
                               message)
                                                   
    d.run
    d.destroy
  end

  def warning_box(message)
    d = Gtk::MessageDialog.new(self,
                               Gtk::Dialog::MODAL | Gtk::Dialog::DESTROY_WITH_PARENT,
                               Gtk::MessageDialog::WARNING,
                               Gtk::MessageDialog::BUTTONS_CLOSE,
                               message)
                                                   
    d.run
    d.destroy
  end

  def error_box(primary, full)
    message = '<span weight="bold" size="larger">' + primary + "</span>\n\n" + full
    l = Gtk::Label.new
    l.set_markup(message)
    d = Gtk::Dialog.new
    d.set_title(_('Error'))
    hbox = Gtk::HBox.new(false, 5)
    d.vbox.pack_start(hbox)
    image = Gtk::Image.new
    image.set(Gtk::Stock::DIALOG_ERROR, Gtk::IconSize::DIALOG)
    hbox.pack_start(image)
    hbox.pack_start(l)
    d.add_button(Gtk::Stock::CLOSE, 1)
    d.show_all
    d.run
    d.destroy
  end

  def select_row_col(titre)
    d = Gtk::Dialog.new
    d.set_title(titre)

    # Row/column selection
    t = Gtk::Table.new(4,4)
    b = Array.new
    b[0] = Gtk::ToggleButton.new
    t.attach(b[0], 1,3, 0,1)
    b[1] = Gtk::ToggleButton.new
    t.attach(b[1], 3,4, 1,3)
    b[2] = Gtk::ToggleButton.new
    t.attach(b[2], 1,3, 3,4)
    b[3] = Gtk::ToggleButton.new
    t.attach(b[3], 0,1, 1,3)
    t.set_size_request(200,200)

    d.vbox.pack_start(Gtk::Label.new(_('Please select a row or column')),
		      false, true, 0)
    d.vbox.pack_start(t, true, true, 5)
    d.vbox.pack_start(Gtk::Label.new(_("Warning: if you add or remove a row at the top\n") \
                                     << _("your map will be shifted to the left. You may\n") \
                                     << _("want to add a column to the left before.")),
		      false, true, 0)

    # Button box
    bb = Gtk::HButtonBox.new
    ok = Gtk::Button.new(Gtk::Stock::OK)
    ok.signal_connect('clicked') {
      [1, 3, 0, 2].each { |i| yield i if b[i].active? }
      d.hide
    }
    cancel = Gtk::Button.new(Gtk::Stock::CANCEL)
    cancel.signal_connect('clicked') {
      d.hide
    }
    bb.pack_start cancel
    bb.pack_start ok
    
    d.action_area.pack_start(bb, false, true, 0)
    d.show_all
  end

  def reset
    @file = nil
    @title.set_text ''
    @nop.set_value 4
    @vp.set_value 10
    @trade.set_active true
    @stricttrade.set_active false
    @n_roads.set_value 15
    @n_ships.set_value 15
    @n_settl.set_value 5
    @n_city.set_value 4
    @n_bridge.set_value 0
    @n_resource.set_value 19
    @n_rb.set_value 2
    @n_mono.set_value 2
    @n_plenty.set_value 2
    @n_chapel.set_value 1
    @n_univ.set_value 1
    @n_governor.set_value 1
    @n_library.set_value 1
    @n_market.set_value 1
    @n_soldier.set_value 13
    @rand_terrain.set_active false
    if (@map != nil)
      @map.hex = Array.new
      @map.ns = NoSetupNodes.new
    end
  end

  def open(filename)
    begin
      comment_location = @start_comment.buffer
      f = File.new(filename, 'r')
      @wd = File.dirname(filename) << '/'
      reset
      @file = filename
      chits = Array.new
      hexes = Array.new
      max = 0
      while ( line = f.gets )
        case line
        when /^title\s+(.+)/
          comment_location = @middle_comment.buffer
          @title.set_text $1
        when /^num-players\s+(\d+)/
          @nop.set_value $1.to_i
        when /^victory-points\s+(\d+)/
          @vp.set_value $1.to_i
        when /^domestic-trade/
          @trade.set_active true
        when /^strict-trade/
          @stricttrade.set_active true
        when /^random-terrain/
          @rand_terrain.set_active true
        when /^num-roads\s+(\d+)/
          @n_roads.set_value $1.to_i
        when /^num-ships\s+(\d+)/
          @n_ships.set_value $1.to_i
        when /^num-settlements\s+(\d+)/
          @n_settl.set_value $1.to_i
        when /^num-cities\s+(\d+)/
          @n_city.set_value $1.to_i
        when /^num-bridges\s+(\d+)/
          @n_bridge.set_value $1.to_i
        when /^resource-count\s+(\d+)/
          @n_resource.set_value $1.to_i
        when /^develop-road\s+(\d+)/
          @n_rb.set_value $1.to_i
        when /^develop-monopoly\s+(\d+)/
          @n_mono.set_value $1.to_i
        when /^develop-plenty\s+(\d+)/
          @n_plenty.set_value $1.to_i	
        when /^develop-chapel\s+(\d+)/
          @n_chapel.set_value $1.to_i
        when /^develop-university\s+(\d+)/
          @n_univ.set_value $1.to_i
        when /^develop-governor\s+(\d+)/
          @n_governor.set_value $1.to_i
        when /^develop-library\s+(\d+)/
          @n_library.set_value $1.to_i
        when /^develop-market\s+(\d+)/
          @n_market.set_value $1.to_i
        when /^develop-soldier\s+(\d+)/
          @n_soldier.set_value $1.to_i
        when /^use-pirate/
          @use_pirate.set_active true
        when /^chits (.*)/
          chits = $1.split ','
        when /^nosetup\s+(\d+)\s+(\d+)\s+(\d+)/
          @map.ns.add( [$1, $2, $3] )
        when /^map/
          i = 0
          while ( line = f.gets and not line === ".\n" )
            @map.hex[i] = Array.new
            j = 0
            a = line.split ','
            a.each { |item|
              if item =~ /\s*([tfhmgpd\-]|s[lbowg\?]?)\s*(\d*)(\+?)\s*/
                id = $2
                type = $1
                random = $3 == ''
                if type === 't'
                  h = 'lumber'
                elsif type === 'f'
                  h = 'grain'
                elsif type === 'h'
                  h = 'brick'
                elsif type === 'm'
                  h = 'ore'
                elsif type === 'p'
                  h = 'wool'
                elsif type === 'd'
                  h = 'desert'
                elsif type === 's'
                  h = 'sea'
                elsif type === 'g'
                  h = 'gold'
                elsif type === '-'
                  h = 'void'
                elsif type =~ /^s[?ogwlb]$/
                  h = type
                else
                  print "Unknown type: " , type , "\n"
                end
                my_hex = Hex.new(h)
                if not my_hex.no_chit? then my_hex.shuffle = random end

                @map.hex[i][j] = my_hex

                if (id != nil) and (id =~ /\d+/)
                  if (type =~ /^s/)
                    if (id =~ /[0-5]/)
                      my_hex.number = id.to_i
                    else
                      print "Error with port number.\n"
                    end
                  else
                    hexes[id.to_i] = my_hex
                  end
                else
                  if (type =~ /^s/)
                    my_hex.number = 0
                  end
                end

                j=j+1
              end
            }
            max = j if j > max
            i = i+1
          end
          comment_location = @end_comment.buffer
        else
          s = comment_location.get_text
          s << line
          comment_location.set_text(s)
        end
      end
      f.close

      # Update width and height
      @map.width = max
      @map.height = @map.hex.length

      # Pad with 'void' hexes if needed
      @map.hex.each { |row|
        d = max - row.length
        if (d > 0)
          0.upto(d-1) {
            row.push Hex.new('void')
          }
        end
      }

      # Attribute chits to hexes
      i = 0
      hexes.each { |h|
        if (h != nil) and (h.type != 'desert') and
            (h.type != nil) and (not h.type =~ /^s/)
          h.number = chits.shift.to_i
        end
        i = i + 1
      }

      @map.repaint
      @map.redisplay
      set_title File.basename(@file) + " - Camato"
    rescue
      error_box(_('Error during opening'), $!)
    end
  end

  def new_map(nr, nc)
    reset
    @map.hex = Array.new
    0.upto(nr-1) { |i|
      @map.hex[i] = Array.new
      0.upto(nc-1) { |j|
	@map.hex[i][j] = Hex.new 'sea'
      }
    }
    @map.width = nc
    @map.height = nr
    @map.repaint
    @map.redisplay
    set_title _("Untitled - Camato") 
  end

  def save(filename)

    # Check for some warnings
    msg = ''

    if @title.text === ''
      msg << _("- You didn't give a title to your map\n")
    end

    if (not filename =~ /.game$/)
      msg << _("- In order to be recognised by the gnocatan\n") \
      << _("server, your filename must end by .game\n")
    end

    has_chits = false
    @map.hex.each { |row|
      row.each { |hex|
        if (hex.number != nil)
          has_chits = true
          break
        end
      }
    }

    if (not has_chits)
      msg << _("- You didn't set any numbers on the hexagons.\nTo do that,") \
      << _(" press a number when your cursor is over an hexagon.\n")
    end

    warning_box(msg) if msg != ''

    begin
      # Save
      f = File.new(filename, 'w')
      f.print(@start_comment.buffer.get_text)

      @wd = File.dirname(filename) << '/'
      f.print 'title ' , @title.text , "\n" ,
              'num-players ' , @nop.text , "\n" ,
                   'victory-points ' , @vp.text , "\n"
      f.print "domestic-trade\n" if @trade.active?
      f.print "strict-trade\n" if @stricttrade.active?
      f.print "random-terrain\n" if @rand_terrain.active?
      f.print 'num-roads ' , @n_roads.text , "\n" ,
              'num-ships ' , @n_ships.text , "\n" ,
                   'num-settlements ' , @n_settl.text , "\n" ,
                        'num-cities ' , @n_city.text , "\n" ,
                             'num-bridges ' , @n_bridge.text , "\n" ,
                                  'resource-count ' , @n_resource.text , "\n"
      f.print 'develop-road ' , @n_rb.text , "\n" ,
              'develop-monopoly ' , @n_mono.text , "\n" ,
                       'develop-plenty ' , @n_plenty.text , "\n" ,
                                'develop-chapel ' , @n_chapel.text , "\n" ,
                                         'develop-university ' , @n_univ.text , "\n" ,
                                                  'develop-governor ' , @n_governor.text , "\n"
      f.print 'develop-library ' , @n_library.text , "\n" ,
              'develop-market ' , @n_market.text , "\n" ,
                       'develop-soldier ' , @n_soldier.text , "\n"
      f.print "use-pirate\n" if @use_pirate.active?
      f.print(@middle_comment.buffer.get_text)

      chits = "chits "
      map_str = "map\n"
      i = 0
      @map.hex.each { |row|
        row.each { |h|
          if h.type == 'sea'
            map_str << 's   ,'
          elsif h.type == 'void'
            map_str << '-   ,'
          elsif h.type == 'desert'
            map_str << 'd' << format("%-3d", i) << ','
            i = i+1
          elsif h.port?
            map_str << h.type
            map_str << h.number.to_s
            map_str << ' ,'
          else
            if h.type == 'lumber'
              t = 't'
            elsif h.type == 'grain'
              t = 'f'
            elsif h.type == 'brick'
              t = 'h'
            elsif h.type == 'ore'
              t = 'm'
            elsif h.type == 'wool'
              t = 'p'
            elsif h.type == 'gold'
              t = 'g'
            end
            map_str << t << format("%d", i)
            if h.not_random? then map_str << '+' end
            map_str << ' ,'
            if (not h.number == nil)
              chits << h.number.to_s
              chits << ","
            end
            i = i + 1
          end
        }
        map_str << "\n"
      }
      f.print chits , "\n"
      f.print map_str , ".\n"
      @map.ns.each { |n| f.print "nosetup ", n.gsub(',',' '), "\n" }
      f.print(@end_comment.buffer.get_text)
      f.close
      set_title File.basename(@file) + " - Camato"
    rescue
      error_box(_('Error during saving'), $!)
    end
  end

  def save_as
    fs = Gtk::FileSelection.new(_('Save as'))
    fs.ok_button.signal_connect('clicked') {
      fs.hide
      @file = fs.filename
      save @file
    }
    fs.cancel_button.signal_connect('clicked') {
      fs.hide
    }
    fs.show
  end

end
