class Map < Canvas

  attr_reader :width, :height, :hex, :cols, :ns
  attr_writer :width, :height, :hex, :ns

  def initialize(width, height)
    super()
    @width = width
    @height = height
    @x = 0
    @y = 0
    # Init the hexagons array
    @hex = Array.new
    0.upto(height-1) { |i|
      @hex[i] = Array.new
      0.upto(width-1) { |j| @hex[i][j] = Hex.new 'sea' }
    }

    # Init the no-setup nodes array
    @ns = NoSetupNodes.new

    set_events Gdk::Event::BUTTON_PRESS_MASK | Gdk::Event::KEY_PRESS_MASK |
         Gdk::Event::ENTER_NOTIFY_MASK |  Gdk::Event::POINTER_MOTION_MASK

    set_flags Gtk::Widget::CAN_FOCUS
    signal_connect('enter_notify_event') { grab_focus }
    signal_connect('motion_notify_event') { |w, e|
      @x = e.x
      @y = e.y
    }
    signal_connect('key_press_event') { |w, e|
      # hack
      my_str = "x"
      i = Gdk::Keyval.to_unicode(e.keyval)
      if i < 0 or i > 255 then i = 0 end
      my_str[0] = i
      set_resource(my_str)
    }
    set_size_request(550, 480)
  end

  def redisplay
    g = window.geometry
    queue_draw_area(0, 0, g[2], g[3])
  end

  def repaint
    if cols == nil
      @l = create_pango_layout
      # initialize colors
      @cols = Hash.new

      c = Gdk::Color.new(0xffff, 0xffff, 0xffff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['white'] = Gdk::GC.new window
      @cols['white'].set_foreground(c)

      c = Gdk::Color.new(0xafff, 0xd3ff, 0xffff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['sea'] = Gdk::GC.new window
      @cols['sea'].set_foreground(c)

      c = Gdk::Color.new(0xc3ff, 0xffff, 0xa5ff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['wool'] = Gdk::GC.new window
      @cols['wool'].set_foreground(c)

      c = Gdk::Color.new(0xddff, 0xabff, 0x71ff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['brick'] = Gdk::GC.new window
      @cols['brick'].set_foreground(c)

      c = Gdk::Color.new(0x69ff, 0xb5ff, 0x6bff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['lumber'] = Gdk::GC.new window
      @cols['lumber'].set_foreground(c)

      c = Gdk::Color.new(0xeaff, 0xeaff, 0x8fff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['grain'] = Gdk::GC.new window
      @cols['grain'].set_foreground(c)

      c = Gdk::Color.new(0xc9ff, 0xc9ff, 0xc9ff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['ore'] = Gdk::GC.new window
      @cols['ore'].set_foreground(c)

      c = Gdk::Color.new(0xf9ff, 0xebff, 0xb1ff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['desert'] = Gdk::GC.new window
      @cols['desert'].set_foreground(c)

      c = Gdk::Color.new(0xd8ff, 0xcaff, 0xb6ff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['void'] = Gdk::GC.new window
      @cols['void'].set_foreground(c)

      c = Gdk::Color.new(0xffff, 0xd0ff, 0x16ff)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['gold'] = Gdk::GC.new window
      @cols['gold'].set_foreground(c)

      c = Gdk::Color.new(0xffff, 0x0000, 0x0000)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @cols['nosetup'] = Gdk::GC.new window
      @cols['nosetup'].set_foreground(c)
    end

    # Compute the width/height of an hexagon
    g = window.geometry
    m1 = Float(g[2])/(@width+2)
    m2 = Float(g[3])/(@height+1)
    if m1 < m2 then m = m1
    else m = m2
    end

    # hs and vs are the horizontal and vertical steps
    # used to draw the hexagons. You start at one point
    # then move one step vertically, then one step
    # vertically and horizontally, and so on.
    hs = m/2
    vs = m/3

    # Clear the drawing area
    @buffer.draw_rectangle(style.white_gc, true,
			   0,0,g[2],g[3]);

    # Draw the hexagons
    0.upto(@height-1) { |i|
      0.upto(@width-1) { |j|

	# Computes (cx, cy), the center of the hexagon
	cx = cy = 0
	if (i.modulo(2) == 0)
	  cx = (j+1)*m
	  cy = (i+1)*m
	else
	  cx = (j+1)*m + hs
	  cy = (i+1)*m
	end

	# Computes the coordinates of each edge
	hex = @hex[i][j]
	edges = Array.new
	edges.push [ cx + hs, cy + vs ]
	edges.push [ cx + hs, cy - vs ]
	edges.push [ cx, cy - 2*vs ]
	edges.push [ cx - hs, cy - vs ]
	edges.push [ cx - hs, cy + vs ]
	edges.push [ cx , cy + 2*vs ]
	edges.push [ cx + hs, cy + vs ]

	if hex.type != nil
	  # find the color for this hexagon.
	  gc = @cols[hex.type]
	  # if none, then use the default.
	  if gc == nil then gc = @cols['white'] end

	  if hex.port?

	    @buffer.draw_polygon(@cols['sea'], true, edges)

	    # Compute the orientation
	    x1 = 0
	    x2 = 0
	    y1 = 0
	    y2 = 0
	    if hex.number == 0
	      # east
	      x1 = cx + hs
	      y1 = cy - vs
	      x2 = cx + hs
	      y2 = cy + vs
	    elsif hex.number == 1
	      # north-east
	      x1 = cx + hs
	      y1 = cy - vs
	      x2 = cx
	      y2 = cy - 2*vs
	    elsif hex.number == 2
	      # north-west
	      x1 = cx - hs
	      y1 = cy - vs
	      x2 = cx
	      y2 = cy - 2*vs
	    elsif hex.number == 3
	      # west
	      x1 = cx - hs
	      y1 = cy - vs
	      x2 = cx - hs
	      y2 = cy + vs
	    elsif hex.number == 4
	      # south-west
	      x1 = cx - hs
	      y1 = cy + vs
	      x2 = cx
	      y2 = cy + 2*vs
	    elsif hex.number == 5
	      # south-east
	      x1 = cx + hs
	      y1 = cy + vs
	      x2 = cx
	      y2 = cy + 2*vs 
	    end

	    @buffer.draw_line(style.black_gc, cx, cy, x1, y1)
	    @buffer.draw_line(style.black_gc, cx, cy, x2, y2)

	    if hex.type === 's?'
	      text = '3:1'
	    else
	      text = hex.type[1,1].to_s.upcase
	    end

            @l.set_text(' 3:1 ')
            s = @l.pixel_size
	    textw = s[0]
	    texth = textw
	    @buffer.draw_arc(style.white_gc, true,
			     cx - textw/2, cy - texth/2,
			     textw, texth,
			     0, 360*64)
	    @buffer.draw_arc(style.black_gc, false,
			     cx - textw/2, cy - texth/2,
			     textw, texth,
			     0, 360*64)

            @l.set_text(text)
            s = @l.pixel_size
	    @buffer.draw_layout(style.black_gc, cx - s[0]/2, cy - texth/3, @l)
	  else
	    @buffer.draw_polygon(gc, true, edges)
	    if (hex.no_chit? and $labels == 2) or
		((not hex.no_chit?) and $labels >= 1)
              @l.set_text(_(hex.type))
              s = @l.pixel_size
	      @buffer.draw_layout(style.black_gc,
				cx - s[0]/2,
				cy - s[1],
				@l) 
	    end
            if hex.not_random?
              @l.set_text('x')
              s = @l.pixel_size
              @buffer.draw_layout(style.black_gc,
                                  cx - s[0]/2,
                                  cy + s[1]/2,
                                  @l)
            end
            if hex.number != nil
              @l.set_text(hex.number.to_s)
              s = @l.pixel_size
              @buffer.draw_layout(style.black_gc,
                                cx - s[0]/2,
                                cy,
                                @l)
            end
	  end
	end

	old = nil
	edges.each { |point|
	  if (old.nil?)
	    old = point
	  else
	    @buffer.draw_line(style.black_gc, old[0], old[1],
			      point[0], point[1])
	    old = point
	  end
	}
      }
    }

    # Draw the no-setup nodes
    @ns.each { |string|
      node = string.split(',')

      # Find the hexagon we're in
      i = node[1].to_i
      j = node[0].to_i
      cx = cy = 0
      if (i.modulo(2) == 0)
        cx = (j+1)*m
        cy = (i+1)*m
      else
        cx = (j+1)*m + hs
        cy = (i+1)*m
      end

      # Now find the position of the node on that hexagon
      case node[2].to_i
      when 0
        cx = cx + hs
        cy = cy - vs
      when 1
        cy = cy - 2*vs
      when 2
        cx = cx - hs
        cy = cy - vs
      when 3
        cx = cx - hs
        cy = cy + vs
      when 4
        cy = cy + 2*vs
      when 5
        cx = cx + hs
        cy = cy + vs
      end

      # Draw the node
      @buffer.draw_arc(@cols['nosetup'], true,
                       cx - 5, cy - 5,
                       10, 10,
                       0, 360*64)
    }
  end

  def configure_event(w, e)
    super
    repaint
  end

  # Called when the user clicks on the map. See Camato class.
  def click(tool)
    case tool
    when 'rotate'
      set_resource(' ')
    when 'nosetup'
      set_resource('n')
    when 'ns_hex'
      set_resource('h')
    when 'no_shuffle'
      set_resource('r')
    when '10'
      set_resource('1')
      set_resource('0')
    when '11'
      set_resource('1')
      set_resource('1')
    when '12'
      set_resource('1')
      set_resource('2')
    else
      set_resource(tool)
    end
  end

  # Sets the resource at coordinatres (x,y) to the given type.
  # Tool can be the first letter of the resource, or 'p'ort, or
  # a number, or a space (port rotation)...
  def set_resource(tool)
    g = window.geometry

    m1 = Float(g[2])/(@width+2)
    m2 = Float(g[3])/(@height+1)
    if m1 < m2 then m = m1
    else m = m2
    end

    row = (@y / m).round - 1
    if (row.modulo(2) == 0) 
      col = (@x / m).round - 1
    else
      col = ((@x-m/2) / m).round - 1
    end

    # Computes (cx, cy), the center of the hexagon
    hs = m/2
    vs = m/3
    cx = cy = 0
    if (row.modulo(2) == 0)
      cx = (col+1)*m
      cy = (row+1)*m
    else
      cx = (col+1)*m + hs
      cy = (row+1)*m
    end

    return if (col == nil or row == nil or
	       col < 0 or col >= @width or
	       row < 0 or row >= @height)

    h = @hex[row][col]
    case tool
    when 'b'
      h.type = (h.port?) ? 'sb' : 'brick'
    when 'o'
      h.type = (h.port?) ? 'so' : 'ore'
    when 'w'
      h.type = (h.port?) ? 'sw' : 'wool'
    when 'g'
      h.type = (h.port?) ? 'sg' : 'grain'
    when 'l'
      h.type = (h.port?) ? 'sl' : 'lumber'
    when 'd'
      h.type = 'desert'
    when 's'
      h.type = 'sea'
    when 'v'
      h.type = 'void'
    when 'a'
      h.type = 'gold'
    when ' '
      if h.port? then h.rotate end
    when 'n'
      a = [col, row]
      # Get the node we've clicked on
      pos = -1
      if ( (@x - cx).abs <= 10 )
        if (@y < cy - 2*vs + 10)
          pos = 1
        elsif (@y > cy + 2*vs - 10)
          pos = 4
        end
      elsif ( (@y - (cy-vs)).abs <= 10 )
        if (@x < cx - hs + 10)
          pos = 2
        elsif (@x > cx + hs - 10)
          pos = 0
        end
      elsif ( (@y - (cy+vs)).abs <= 10 )
        if (@x < cx - hs + 10)
          pos = 3
        elsif (@x > cx + hs - 10)
          pos = 5
        end
      end
      if (pos != -1)
        a[2] = pos
        if (@ns.has?(a))
          @ns.remove(a)
        else
          @ns.add(a)
        end
      end
    when 'h'
      is_nosetup = true
      a = Array.new
      0.upto(5) { |i|
        a[i] = [col, row, i]
        if (not @ns.has?(a[i])) then is_nosetup = false end
      }
      if (is_nosetup)
        a.each { |s| @ns.remove(s) }
      else
        a.each { |s| @ns.add(s) }
      end
    when 'r'
      if $use_cvs then
        h.shuffle = h.not_random? unless h.no_chit?
      end
    when 'p'
      h.type = 's?'
      if h.number == nil then h.number = 0 end
    when /^s./
      h.type = tool
      if h.number == nil then h.number = 0 end
    when /([3-9])/
      h.number = $1.to_i unless h.no_chit?
    when /[0-2]/
      if (not h.no_chit?)
        if (@old_num == 1)
          h.number = tool.to_i + 10
          @old_num = 0
        elsif (tool.to_i == 2)
          h.number = 2
        elsif (tool.to_i == 1) 
          @old_num = 1
        end
      end
    end

    repaint
    queue_draw_area(@x - 2*m, @y - 2*m, 4*m, 4*m)
  end

end
