# This class represents the nodes where one can't place
# a building during the setup phase. In gnocatan, these
# are triplets [row, column, position], where row and
# columns identify an hex, and position is the node on
# that hex. Since there can be multiple triplets that
# represent the same node, we have to use this class
# to avoid duplication. We keep the left-most, upper-most
# hex for each node.
# For nodes that are on the left or the top borders,
# this is not possible, so we keep the right-most, bottom-
# most hexagons for those nodes.
class NoSetupNodes
  def initialize
    @ns = Hash.new
  end

  def add(node)
    @ns[min(node)] = true
  end

  def remove(node)
    @ns.delete( min(node))
  end

  def remove_all
    @ns.clear
  end

  def has?(node)
    return @ns.has_key?( min(node) )
  end

  def min(node)
    n = [ node[0], node[1], node[2] ]
    case n[2]
    when 0
      if (n[1] != 0)
        n[1] = n[1] - 1
        n[0] = n[0] + 1 if (n[1].modulo(2) == 0)
        n[2] = 4
      end
    when 1
      if (n[1] != 0) and (n[0] != 0)
        n[1] = n[1] - 1
        n[0] = n[0] - 1 if (n[1].modulo(2) == 1)
        n[2] = 5
      end
    when 2
      if (n[1] != 0) and (n[0] != 0)
        n[1] = n[1] - 1
        n[0] = n[0] - 1 if (n[1].modulo(2) == 1)
        n[2] = 4
      end
    when 3
      if (n[1] != 0) and (n[0] != 0)
        n[0] = n[0] - 1
        n[2] = 5
      end
    end
    return n.join(',')
  end

  def each
    @ns.each_key { |k| yield k }
  end
end
