# Graphs
class Graph < Canvas

  def initialize()
    super
    @values = Hash.new
    @l = create_pango_layout
  end

  def repaint

    if @red == nil
      c = Gdk::Color.new(0xffff, 0x0000, 0x0000)
      Gdk::Colormap.system.alloc_color(c, true, true)
      @red = Gdk::GC.new window
      @red.set_foreground(c)
    end

    g = window.geometry
    @buffer.draw_rectangle(style.white_gc, true,
			   0,0,g[2],g[3]);
    width = g[2]/(@values.length+1)

    # computes the sum of all values
    total = 0
    @values.each { |key, val|
      total = total + val
    }

    return if total == 0

    # computes the percentages
    p = Hash.new
    sa = @values.sort
    max_p = 0
    sa.each { |name, val|
      p[name] = (val * 100) / total
      if (p[name] > max_p) then max_p = p[name] end
    }

    i = 1;
    p.sort.each { |name, val|

      n = name.to_s
      @l.set_text(n)
      s = @l.pixel_size

      # Print x-axis legend
      x = i * width
      y = g[3] - s[1] - 5

      @buffer.draw_layout(style.black_gc, x, y, @l)

      # Print value
      r_x = x
      r_height = (y-50) * val/max_p
      r_y = y - r_height - s[1] - 5
      r_width = width/3
      @buffer.draw_rectangle(@red, true, r_x, r_y, r_width, r_height)

      i = i+1
    }

    # print y-axis legend
    vals = [ 0, 0.25, 0.5, 0.75, 1 ]
    vals.each { |n|
      str = "#{n*max_p} %"
      @l.set_text(str)
      s = @l.pixel_size

      y = g[3] - s[1] - 5
      r_height = (y-50) * n
      r_y = y - r_height - s[1] - 5
      r_x = 5
      @buffer.draw_layout(style.black_gc, r_x, r_y - 2, @l)
      @buffer.draw_line(style.black_gc, r_x, r_y,
                        r_x + g[2],
                        r_y)
    }
  end

  def add(name, val)
    value = @values[name]
    if value == nil
      @values[name] = val
    else
      value = value + val
      @values[name] = value
    end
  end

  def configure_event(w, e)
    super
    repaint
  end
end

# Statistics
class Stats < Gtk::Window

  def create_graphs
    # Create the graphs
    g = Graph.new
    @map.hex.each { |array|
      2.upto(12) { |i| g.add(i, 0) unless i == 7}
      array.each { |h|
	g.add(h.number, 1) unless h.no_chit? or h.number == nil
      }
    }
    @res['All (by chits)'] = g

    g = Graph.new
    @map.hex.each { |array|
      array.each { |h|
	g.add(_(h.type), 1) unless h.no_chit?
      }
    }
    @res['All (by type)'] = g

    ['Brick', 'Grain', 'Gold', 'Lumber',
      'Ore', 'Wool'].each { |string|
      g = Graph.new
      @map.hex.each { |array|
	2.upto(12) { |i| g.add(i, 0) unless i == 7}
	array.each { |h|
	  if (h.type === string.downcase)
	    g.add(h.number, 1) unless h.number == nil
	  end
	}
      }
      @res[string] = g
    }

    @res['Resources'] = @res['All (by type)']

    g = Graph.new
    [@app.n_chapel, @app.n_univ, @app.n_governor, @app.n_library,
      @app.n_market].each { |entry|
      g.add(_('Vict. points'), entry.text.to_i)
    }
    g.add(_('Soldiers'), @app.n_soldier.text.to_i)
    g.add(_('Road build.'), @app.n_rb.text.to_i)
    g.add(_('Monopolies'), @app.n_mono.text.to_i)
    g.add(_('Yr.of plenty'), @app.n_plenty.text.to_i)
    @res['Dev cards'] = g
  end

  def initialize(app)
    super Gtk::Window::Type::TOPLEVEL
    set_title(_('Statistics'))
    @res = Hash.new
    @app = app
    @map = app.map

    @vbox = Gtk::VBox.new(false, 0)
    add(@vbox)

    hbox = Gtk::HBox.new(false, 0)
    @vbox.pack_start(hbox, true, true, 0)

    create_graphs

    tree = Gtk::TreeStore.new(String, String)
    i = tree.append(nil)
    i[0] = _('Resources')
    i[1] = 'Resources'
    i2 = tree.append(i)
    i2[0] = _('All (by type)')
    i2[1] = 'All (by type)'
    i2 = tree.append(i)
    i2[0] = _('All (by chits)')
    i2[1] = 'All (by chits)'
    i2 = tree.append(i)
    i2[0] = _('Brick')
    i2[1] = 'Brick'
    i2 = tree.append(i)
    i2[0] = _('Gold')
    i2[1] = 'Gold'
    i2 = tree.append(i)
    i2[0] = _('Grain')
    i2[1] = 'Grain'
    i2 = tree.append(i)
    i2[0] = _('Lumber')
    i2[1] = 'Lumber'
    i2 = tree.append(i)
    i2[0] = _('Ore')
    i2[1] = 'Ore'
    i2 = tree.append(i)
    i2[0] = _('Wool')
    i2[1] = 'Wool'
    i = tree.append(nil)
    i[0] = _('Properties')
    i[1] = 'Properties'
    i2 = tree.append(i)
    i2[0] = _('Dev cards')
    i2[1] = 'Dev cards'
    view = Gtk::TreeView.new(tree)
    renderer = Gtk::CellRendererText.new()
    col = Gtk::TreeViewColumn.new(_('Section'), renderer, :text => 0)
    view.append_column(col)
    view.expand_all
    select = view.selection
    select.set_mode(Gtk::SELECTION_SINGLE)
    select.signal_connect('changed') { |sel|
      if (not @res[sel.selected[1]] == nil)
        hbox.remove @current_res unless @current_res == nil
        @current_res = @res[sel.selected[1]]
        hbox.pack_start(@current_res, true, true, 0)
        @current_res.show
      end
    }

    s = Gtk::ScrolledWindow.new
    s.add(view)
    s.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    s.set_size_request(175, 50)
    hbox.pack_start(s, false, true, 0)

    bbox = Gtk::HButtonBox.new
    bbox.spacing=10
    bbox.set_border_width(5)
    bbox.layout_style = Gtk::ButtonBox::END
    update = Gtk::Button.new(Gtk::Stock::REFRESH)
    update.signal_connect('clicked') {
      create_graphs
      hbox.remove @current_res
      @current_res = @res[view.selection.selected[1]]
      hbox.pack_start(@current_res, true, true, 0)
      @current_res.show
    }
    bbox.pack_start update
    close = Gtk::Button.new(Gtk::Stock::CLOSE)
    close.signal_connect('clicked') {
      destroy
    }
    bbox.pack_start close
    @vbox.pack_start(bbox, false, true, 5)
    set_size_request(600, 300)
  end
end
